from mltrace.db.base import Base
from mltrace.db.models import ComponentRun, PointerTypeEnum
from sqlalchemy import create_engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.schema import (
    DropConstraint,
    DropTable,
    MetaData,
    Table,
    ForeignKeyConstraint,
)

import pprint
import sqlalchemy


def _create_engine_wrapper(uri: str, max_retries=5) -> sqlalchemy.engine.base.Engine:
    """Creates engine using sqlalchemy API. Includes max retries parameter."""
    retries = 0
    while retries < max_retries:
        try:
            engine = create_engine(uri)
            return engine
        except Exception as e:
            print(f'DB could not be created with exception {e}. Trying again.')
        retries += 1
    raise RuntimeError('Max retries hit.')


def _initialize_db_tables(engine: sqlalchemy.engine.base.Engine):
    """Initializes tables using sqlalchemy API."""
    Base.metadata.create_all(engine)


def _drop_everything(engine: sqlalchemy.engine.base.Engine):
    """(On a live db) drops all foreign key constraints before dropping all tables.
    Workaround for SQLAlchemy not doing DROP ## CASCADE for drop_all()
    (https://github.com/pallets/flask-sqlalchemy/issues/722)
    """

    con = engine.connect()
    trans = con.begin()
    inspector = Inspector.from_engine(engine)

    # We need to re-create a minimal metadata with only the required things to
    # successfully emit drop constraints and tables commands for postgres (based
    # on the actual schema of the running instance)
    meta = MetaData()
    tables = []
    all_fkeys = []

    for table_name in inspector.get_table_names():
        fkeys = []

        for fkey in inspector.get_foreign_keys(table_name):
            if not fkey["name"]:
                continue

            fkeys.append(ForeignKeyConstraint((), (), name=fkey["name"]))

        tables.append(Table(table_name, meta, *fkeys))
        all_fkeys.extend(fkeys)

    for fkey in all_fkeys:
        con.execute(DropConstraint(fkey))

    for table in tables:
        con.execute(DropTable(table))

    trans.commit()
    Base.metadata.drop_all(engine)


def _map_extension_to_enum(filename: str) -> PointerTypeEnum:
    """Infers the relevnat enum for the filename."""
    data_extensions = ['csv', 'pq', 'parquet',
                       'txt', 'md', 'rtf', 'tsv', 'xml', 'pdf']
    model_extensions = ['h5', 'hdf5', 'joblib',
                        'pkl', 'pickle', 'ckpt', 'mlmodel']

    words = filename.split('.')

    if len(words) < 1:
        return PointerTypeEnum.UNKNOWN

    extension = words[-1].lower()

    if extension in data_extensions:
        return PointerTypeEnum.DATA

    if extension in model_extensions:
        return PointerTypeEnum.MODEL

    # TODO(shreyashankar): figure out how to handle output id
    return PointerTypeEnum.UNKNOWN



from __future__ import annotations
from datetime import datetime
from mltrace.db.base import Base
from sqlalchemy import Column, String, LargeBinary, Integer, DateTime, Table, ForeignKey, Enum
from sqlalchemy.orm import relationship

import enum
import typing


class PointerTypeEnum(str, enum.Enum):
    DATA = 'DATA'
    MODEL = 'MODEL'
    ENDPOINT = 'ENDPOINT'
    UNKNOWN = 'UNKNOWN'


component_tag_association = Table(
    'component_tags', Base.metadata,
    Column('component_name', String, ForeignKey('components.name')),
    Column('tag_name', String, ForeignKey('tags.name'))
)


class Component(Base):
    __tablename__ = 'components'

    name = Column(String, primary_key=True)
    description = Column(String)
    owner = Column(String)
    component_runs = relationship("ComponentRun", cascade='all, delete-orphan')
    tags = relationship(
        "Tag", secondary=component_tag_association, cascade='all')

    def __init__(self, name: str, description: str, owner: str, tags: typing.List[Tag] = []):
        self.name = name
        self.description = description
        self.owner = owner
        self.tags = []

    def add_tags(self, tags: typing.List[Tag]):
        self.tags += tags


class Tag(Base):
    __tablename__ = 'tags'

    name = Column(String, primary_key=True)

    def __init__(self, name: str):
        self.name = name


class IOPointer(Base):
    __tablename__ = 'io_pointers'

    name = Column(String, primary_key=True)
    pointer_type = Column(Enum(PointerTypeEnum))

    def __init__(self, name: str, pointer_type: PointerTypeEnum = PointerTypeEnum.UNKNOWN):
        self.name = name
        self.pointer_type = pointer_type

    def set_pointer_type(self, pointer_type: PointerTypeEnum):
        self.pointer_type = pointer_type


component_run_input_association = Table(
    'component_runs_inputs', Base.metadata,
    Column('input_path_name', String, ForeignKey('io_pointers.name')),
    Column('component_run_id', Integer, ForeignKey('component_runs.id'))
)

component_run_output_association = Table(
    'component_runs_outputs', Base.metadata,
    Column('output_path_name', String, ForeignKey('io_pointers.name')),
    Column('component_run_id', Integer, ForeignKey('component_runs.id'))
)

component_run_dependencies = Table(
    'component_run_dependencies', Base.metadata,
    Column('component_run_id', Integer, ForeignKey(
        'component_runs.id'), primary_key=True),
    Column('depends_on_component_run_id', Integer, ForeignKey(
        'component_runs.id'), primary_key=True)
)


class ComponentRun(Base):
    __tablename__ = 'component_runs'

    id = Column(Integer, primary_key=True)
    component_name = Column(String, ForeignKey('components.name'))
    git_hash = Column(String)
    code_snapshot = Column(LargeBinary)
    start_timestamp = Column(DateTime)
    end_timestamp = Column(DateTime)
    inputs = relationship(
        "IOPointer", secondary=component_run_input_association, cascade='all')
    outputs = relationship(
        "IOPointer", secondary=component_run_output_association, cascade='all')
    dependencies = relationship('ComponentRun', secondary=component_run_dependencies, primaryjoin=id == component_run_dependencies.c.component_run_id,
                                secondaryjoin=id == component_run_dependencies.c.depends_on_component_run_id, backref="left_component_run_ids", cascade='all')

    def __init__(self, component_name):
        """Initialize ComponentRun, or an instance of a Component's 'run.'"""
        self.component_name = component_name
        self.inputs = []
        self.outputs = []
        self.dependencies = []

    def set_start_timestamp(self, ts: datetime = None):
        """Call this function to set the start timestamp to a specific timestamp or now."""
        self.start_timestamp = ts if ts else datetime.now()

    def set_end_timestamp(self, ts: datetime = None):
        """Call this function to set the end timestamp to a specific timestamp or now."""
        self.end_timestamp = ts if ts else datetime.now()

    def set_code_snapshot(self, code_snapshot: bytes):
        """Code snapshot setter."""
        self.code_snapshot = code_snapshot

    def set_git_hash(self, git_hash: str):
        """Git hash setter."""
        self.git_hash = git_hash

    def add_input(self, input: IOPointer):
        """Add a single input (instance of IOPointer)."""
        self._add_io(input, True)

    def add_inputs(self, inputs: typing.List[IOPointer]):
        """Add a list of inputs (each element should be an instance of IOPointer)."""
        self._add_io(inputs, True)

    def add_output(self, output: IOPointer):
        """"Add a single output (instance of IOPointer)."""
        self._add_io(output, False)

    def add_outputs(self, outputs: typing.List[IOPointer]):
        """Add a list of outputs (each element should be an instance of IOPointer)."""
        self._add_io(outputs, False)

    def _add_io(self, elems: typing.Union[typing.List[IOPointer], IOPointer], input: bool):
        """Helper function to add inputs or outputs."""
        # Elems can be a list or a single IOPointer. Set to a list.
        elems = [elems] if not isinstance(elems, list) else elems
        if input:
            self.inputs += elems
        else:
            self.outputs += elems

    def set_upstream(self, dependencies: typing.Union[typing.List[ComponentRun], ComponentRun]):
        """Set dependencies for this ComponentRun. API similar to Airflow set_upstream."""
        # Dependencies can be a list or a single ComponentRun. Set to a list.
        dependencies = [dependencies] if not isinstance(
            dependencies, list) else dependencies

        self.dependencies += dependencies
        # Drop duplicates
        self.dependencies = list(set(self.dependencies))

    def check_completeness(self) -> dict:
        """Returns a dictionary of success indicator and error messages."""
        status_dict = {'success': True, 'msg': ''}
        if self.start_timestamp is None:
            status_dict['success'] = False
            status_dict['msg'] += f'{self.component_name} ComponentRun has no start timestamp. '
        if self.end_timestamp is None:
            status_dict['success'] = False
            status_dict['msg'] += f'{self.component_name} ComponentRun has no end timestamp. '
        return status_dict



from mltrace.entities.base import Base

import json
import pprint
import typing


class Component(Base):
    """Component abstraction."""

    def __init__(self, name: str, description: str, owner: str, tags: typing.List[str] = []):
        """ Components should have a name, description, and owner. 
        Optionally they will have tags."""
        self._name = name
        self._description = description
        self._owner = owner
        self._tags = tags

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def owner(self) -> str:
        return self._owner

    @property
    def tags(self) -> typing.List[str]:
        return self._tags

    def __repr__(self):
        params = self.to_dictionary()
        return json.dumps(params)



from mltrace.db.models import PointerTypeEnum
from mltrace.entities.base import Base

import json
import pprint


class IOPointer(Base):

    def __init__(self, name: str, pointer_type: PointerTypeEnum = PointerTypeEnum.UNKNOWN):
        self._name = name
        self._pointer_type = pointer_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def pointer_type(self) -> PointerTypeEnum:
        return self._pointer_type

    def __repr__(self):
        params = self.to_dictionary()
        params['pointer_type'] = params['pointer_type'].value
        return json.dumps(params)



from mltrace.db import Store, PointerTypeEnum
from mltrace.entities import Component, ComponentRun, IOPointer

import copy
import functools
import git
import inspect
import logging
import typing
import uuid

_db_uri = 'postgresql://admin:admin@database:5432/sqlalchemy'

# ----------------------- Database management functions ---------------------- #


def set_db_uri(uri: str):
    global _db_uri
    _db_uri = uri


def get_db_uri() -> str:
    global _db_uri
    return _db_uri


def clean_db():
    """Deletes database and reinitializes tables."""
    store = Store(_db_uri, delete_first=True)

# ---------------------------- Creation functions ---------------------------- #


def create_component(name: str, description: str, owner: str, tags: typing.List[str] = []):
    """Creates a component entity in the database."""
    store = Store(_db_uri)
    store.create_component(name, description, owner, tags)


def tag_component(component_name: str, tags: typing.List[str]):
    """Adds tags to existing component."""
    store = Store(_db_uri)
    store.add_tags_to_component(component_name, tags)


def log_component_run(component_run: ComponentRun, set_dependencies_from_inputs=True):
    """Takes client-facing ComponentRun object and logs it to the DB."""
    store = Store(_db_uri)

    # Make dictionary object
    component_run_dict = component_run.to_dictionary()

    component_run_sql = store.initialize_empty_component_run(
        component_run.component_name)

    # Add relevant attributes
    component_run_sql.set_start_timestamp(component_run_dict['start_timestamp'])
    component_run_sql.set_end_timestamp(component_run_dict['end_timestamp'])
    component_run_sql.set_git_hash(component_run_dict['git_hash'])
    component_run_sql.set_code_snapshot(component_run_dict['code_snapshot'])

    # Add I/O
    component_run_sql.add_inputs([store.get_io_pointer(
        inp.name, inp.pointer_type) for inp in component_run_dict['inputs']])
    component_run_sql.add_outputs([store.get_io_pointer(
        out.name, out.pointer_type) for out in component_run_dict['outputs']])

    # Add dependencies if there is flag to automatically set
    if set_dependencies_from_inputs:
        store.set_dependencies_from_inputs(component_run_sql)

    # Add dependencies explicitly stored in the component run
    for dependency in component_run_dict['dependencies']:
        cr = store.get_history(dependency, 1)[0]
        component_run_sql.set_upstream(cr)

    store.commit_component_run(component_run_sql)


def create_random_ids(num_outputs=1) -> typing.List[str]:
    """Returns a list of num_outputs ids that a client can use to tag outputs."""

    return [uuid.uuid4() for _ in range(num_outputs)]

# Log input strings
# function to apply to outputs to log those


def register(component_name: str, inputs: typing.List[str] = [], outputs: typing.List[str] = [], input_vars: typing.List[str] = [], output_vars: typing.List[str] = [], endpoint: bool = False):
    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Construct component run object
            store = Store(_db_uri)
            component_run = store.initialize_empty_component_run(component_name)
            component_run.set_start_timestamp()

            # Add input_vars and output_vars as pointers
            input_pointers = [store.get_io_pointer(
                str(kwargs[var])) for var in input_vars]
            output_pointers = [store.get_io_pointer(str(kwargs[var]), PointerTypeEnum.ENDPOINT) for var in output_vars] if endpoint else [
                store.get_io_pointer(str(kwargs[var])) for var in output_vars]

            # Run function
            value = func(*args, **kwargs)

            # Log relevant info
            component_run.set_end_timestamp()
            input_pointers += [store.get_io_pointer(inp) for inp in inputs]
            output_pointers += [store.get_io_pointer(
                out, PointerTypeEnum.ENDPOINT) for out in outputs] if endpoint else [store.get_io_pointer(out) for out in outputs]
            component_run.add_inputs(input_pointers)
            component_run.add_outputs(output_pointers)
            store.set_dependencies_from_inputs(component_run)

            # Add code versions
            try:
                repo = git.Repo(search_parent_directories=True)
                component_run.set_git_hash(str(repo.head.object.hexsha))
            except:
                logging.info('No git repo found.')

            # Add source code if less than 2^16
            func_source_code = inspect.getsource(func)
            if len(func_source_code) < 2**16:
                component_run.set_code_snapshot(
                    bytes(func_source_code, 'ascii'))

            # Commit component run object to the DB
            store.commit_component_run(component_run)

            return value
        return wrapper
    return actual_decorator


# ------------------------- Basic retrieval functions ------------------------ #


def get_history(component_name: str, limit: int = 10) -> typing.List[ComponentRun]:
    """ Returns a list of ComponentRuns that are part of the component's
    history."""
    store = Store(_db_uri)

    history = store.get_history(component_name, limit)

    # Convert to client-facing ComponentRuns
    component_runs = []
    for cr in history:
        inputs = [IOPointer.from_dictionary(
            iop.__dict__).to_dictionary() for iop in cr.inputs]
        outputs = [IOPointer.from_dictionary(
            iop.__dict__).to_dictionary() for iop in cr.outputs]
        dependencies = [dep.component_name for dep in cr.dependencies]
        d = copy.deepcopy(cr.__dict__)
        d.update({'inputs': inputs, 'outputs': outputs,
                 'dependencies': dependencies})
        component_runs.append(ComponentRun.from_dictionary(d))

    return component_runs


def get_components_with_owner(owner: str) -> typing.List[Component]:
    """ Returns a list of all the components associated with the specified
        order."""
    store = Store(_db_uri)
    res = store.get_components_with_owner(owner)

    # Convert to client-facing Components
    components = []
    for c in res:
        tags = [tag.name for tag in c.tags]
        d = copy.deepcopy(c.__dict__)
        d.update({'tags': tags})
        components.append(Component.from_dictionary(d))

    return components


def get_component_information(component_name: str) -> Component:
    """Returns a Component with the name, info, owner, and tags."""
    store = Store(_db_uri)
    c = store.get_component(component_name)
    tags = [tag.name for tag in c.tags]
    d = copy.deepcopy(c.__dict__)
    d.update({'tags': tags})
    return Component.from_dictionary(d)


def get_component_run_information(component_run_id: str) -> ComponentRun:
    """Returns a ComponentRun object."""
    store = Store(_db_uri)
    cr = store.get_component_run(component_run_id)
    inputs = [IOPointer.from_dictionary(
        iop.__dict__).to_dictionary() for iop in cr.inputs]
    outputs = [IOPointer.from_dictionary(
        iop.__dict__).to_dictionary() for iop in cr.outputs]
    dependencies = [dep.component_name for dep in cr.dependencies]
    d = copy.deepcopy(cr.__dict__)
    if cr.code_snapshot:
        d.update({'code_snapshot': str(cr.code_snapshot.decode('utf-8'))})
    d.update({'inputs': inputs, 'outputs': outputs,
              'dependencies': dependencies})
    return ComponentRun.from_dictionary(d)


def get_components_with_tag(tag: str) -> typing.List[Component]:
    """Returns a list of components with the specified tag."""
    store = Store(_db_uri)
    res = store.get_components_with_tag(tag)

    # Convert to client-facing Components
    components = []
    for c in res:
        tags = [tag.name for tag in c.tags]
        d = copy.deepcopy(c.__dict__)
        d.update({'tags': tags})
        components.append(Component.from_dictionary(d))

    return components


def get_recent_run_ids(limit: int = 50):
    """Returns most recent component run ids."""
    store = Store(_db_uri)
    return store.get_recent_run_ids(limit)


def get_io_pointer(io_pointer_id: str, create=True):
    """Returns IO Pointer metadata."""
    store = Store(_db_uri)
    iop = store.get_io_pointer(io_pointer_id, create)
    return IOPointer.from_dictionary(iop.__dict__)

# ------------------------ Complex retrieval functions ----------------------- #


def backtrace(output_pointer: str):
    """Prints trace for an output id.
        Returns list of tuples (level, ComponentRun) where level is how
        many hops away the node is from the node that produced the output_id."""
    store = Store(_db_uri)
    trace = store.trace(output_pointer)

    # Convert to entities.ComponentRun
    component_runs = []
    for depth, cr in trace:
        inputs = [IOPointer.from_dictionary(iop.__dict__) for iop in cr.inputs]
        outputs = [IOPointer.from_dictionary(
            iop.__dict__) for iop in cr.outputs]
        dependencies = [dep.component_name for dep in cr.dependencies]
        d = copy.deepcopy(cr.__dict__)
        d.update({'inputs': inputs, 'outputs': outputs,
                 'dependencies': dependencies})
        component_runs.append((depth, ComponentRun.from_dictionary(d)))

    return component_runs


def web_trace(output_id: str):
    store = Store(_db_uri)
    return store.web_trace(output_id)



# from __future__ import annotations
from datetime import datetime
from mltrace.db import PointerTypeEnum
from mltrace.db.utils import _map_extension_to_enum
from mltrace.entities.base import Base
from mltrace.entities.io_pointer import IOPointer

import json
import pickle
import pprint
import typing


def get_timestamp() -> int:
    """Returns current timestamp as seconds since epoch."""
    return int(datetime.now().strftime('%s'))


class ComponentRun(Base):
    """Component Run abstraction."""

    def __init__(self, component_name: str, start_timestamp: datetime = None, end_timestamp: datetime = None, inputs: typing.List[IOPointer] = [], outputs: typing.List[IOPointer] = [], git_hash: str = None, code_snapshot: str = None, id: str = None, dependencies: typing.List[str] = []):
        """ Set class attributes. Note that timestamps are represented in
seconds since epoch."""
        self._component_name = component_name
        self._start_timestamp = start_timestamp
        self._end_timestamp = end_timestamp
        self._inputs = inputs
        self._outputs = outputs
        self._git_hash = git_hash
        self._code_snapshot = code_snapshot.decode(
            'utf-8') if isinstance(code_snapshot, (bytes, bytearray)) else code_snapshot
        self._id = id
        self._dependencies = dependencies

    @property
    def component_name(self) -> str:
        return self._component_name

    @property
    def inputs(self) -> typing.List[IOPointer]:
        return self._inputs

    @inputs.setter
    def inputs(self, inputs: typing.List[IOPointer]):
        self._inputs = inputs

    @inputs.deleter
    def inputs(self):
        del self._inputs

    @property
    def outputs(self) -> typing.List[IOPointer]:
        return self._outputs

    @outputs.setter
    def outputs(self, outputs: typing.List[IOPointer]):
        self._outputs = outputs

    @outputs.deleter
    def outputs(self):
        del self._outputs

    @property
    def git_hash(self) -> str:
        return self._git_hash

    @git_hash.setter
    def git_hash(self, new_hash: str):
        self._git_hash = new_hash

    @property
    def code_snapshot(self) -> str:
        return self._code_snapshot

    @code_snapshot.setter
    def code_snapshot(self, new_snapshot: str):
        self._code_snapshot = new_snapshot

    @property
    def start_timestamp(self) -> datetime:
        return self._start_timestamp

    @property
    def end_timestamp(self) -> datetime:
        return self._end_timestamp

    @property
    def dependencies(self) -> typing.List[str]:
        return self._dependencies

    @property
    def id(self) -> str:
        return self._id

    def set_start_timestamp(self, ts=None):
        if ts is None:
            ts = datetime.now()
        self._start_timestamp = ts

    def set_end_timestamp(self, ts=None):
        if ts is None:
            ts = datetime.now()
        self._end_timestamp = ts

    def add_input(self, name: str, pointer_type: PointerTypeEnum = None):
        """Add a single input (instance of IOPointer)."""
        if pointer_type is None:
            pointer_type = _map_extension_to_enum(name)
        self._add_io(IOPointer(name, pointer_type), True)

    def add_inputs(self, inputs: typing.List[IOPointer]):
        """Add a list of inputs (each element should be an instance of IOPointer)."""
        self._add_io(inputs, True)

    def add_output(self, name: str, pointer_type: PointerTypeEnum = None):
        """"Add a single output (instance of IOPointer)."""
        if pointer_type is None:
            pointer_type = _map_extension_to_enum(name)
        self._add_io(IOPointer(name, pointer_type), False)

    def add_outputs(self, outputs: typing.List[IOPointer]):
        """Add a list of outputs (each element should be an instance of IOPointer)."""
        self._add_io(outputs, False)

    def _add_io(self, elems: typing.Union[typing.List[IOPointer], IOPointer], input: bool):
        """Helper function to add inputs or outputs."""
        # Elems can be a list or a single IOPointer. Set to a list.
        elems = [elems] if not isinstance(elems, list) else elems
        if input:
            self.inputs = self.inputs + elems
        else:
            self.outputs = self.outputs + elems

    def set_upstream(self, dependencies: typing.Union[str, typing.List[str]]):
        """ Set dependencies for this ComponentRun. API similar to Airflow
        set_upstream. It will grab the most recent run for the dependency
        name."""
        # Dependencies can be a list or a single string. Set to a list.
        dependencies = [dependencies] if not isinstance(
            dependencies, list) else dependencies

        self._dependencies = self._dependencies + dependencies
        self._dependencies = list(set(self._dependencies))

    def __repr__(self):
        params = self.to_dictionary()
        if params['start_timestamp'] is not None:
            params['start_timestamp'] = params['start_timestamp'].strftime(
                '%Y-%m-%d %l:%M:%S%z')
        if params['end_timestamp'] is not None:
            params['end_timestamp'] = params['end_timestamp'].strftime(
                '%Y-%m-%d %l:%M:%S%z')
        return json.dumps(params)



from flask import Flask, request, Response
from http import HTTPStatus
from mltrace.entities import Component, ComponentRun, IOPointer
from mltrace import get_component_information, get_component_run_information, get_components_with_tag, get_history, web_trace, get_recent_run_ids, get_io_pointer, set_db_uri

import copy
import json
import logging

app = Flask(__name__)
set_db_uri('postgresql://admin:admin@database:5432/sqlalchemy')


def error(err_msg, status_code):
    return Response(json.dumps({"error": err_msg}), status=status_code)


def serialize_component_run(c: Component, cr: ComponentRun) -> str:
    """Serializes component run to display info on a card."""
    web_cr_dict = json.loads(str(cr))

    # Add component information
    web_cr_dict['owner'] = c.owner
    web_cr_dict['description'] = c.description
    web_cr_dict['tags'] = c.tags

    return json.dumps(web_cr_dict)


@app.route('/component_run', methods=['GET'])
def component_run():
    if 'id' not in request.args:
        return error(f'id not specified.', HTTPStatus.NOT_FOUND)

    component_run_id = request.args['id']
    try:
        # Check to make sure the id is actually numeric
        if not component_run_id.isdigit():
            raise RuntimeError()
        component_run = get_component_run_information(component_run_id)
        print(component_run)
        component = get_component_information(component_run.component_name)
        return serialize_component_run(component, component_run)
    except RuntimeError:
        return error(f'Component run ID {component_run_id} not found', HTTPStatus.NOT_FOUND)


@app.route('/io_pointer', methods=['GET'])
def io_pointer():
    if 'id' not in request.args:
        return error(f'id not specified.', HTTPStatus.NOT_FOUND)

    io_pointer_id = request.args['id']
    try:
        res = get_io_pointer(io_pointer_id, create=False)
        return json.dumps(IOPointer.from_dictionary(res.__dict__).to_dictionary())
    except RuntimeError:
        return error(f'IOPointer {io_pointer_id} not found', HTTPStatus.NOT_FOUND)


@app.route('/tag', methods=['GET'])
def tag():
    if 'id' not in request.args:
        return error(f'id not specified.', HTTPStatus.NOT_FOUND)

    tag_name = request.args['id']
    try:
        components = get_components_with_tag(tag_name)
        return str(components)
    except RuntimeError:
        return error(f'Tag {tag_name} not found', HTTPStatus.NOT_FOUND)


@app.route('/history', methods=['GET'])
def history():
    if 'component_name' not in request.args:
        return error(f'component_name not specified.', HTTPStatus.NOT_FOUND)

    component_name = request.args['component_name']
    limit = request.args['limit'] if 'limit' in request.args else None

    try:
        history = get_history(
            component_name, limit) if limit else get_history(component_name)
        return str(history)
    except RuntimeError:
        return error(f'Component {component_name} has no runs', HTTPStatus.NOT_FOUND)


@app.route('/component', methods=['GET'])
def component():
    if 'id' not in request.args:
        return error(f'id not specified.', HTTPStatus.NOT_FOUND)

    component_name = request.args['id']
    try:
        component = get_component_information(component_name)
        return str(component)
    except RuntimeError:
        return error(f'Component {component_name} not found', HTTPStatus.NOT_FOUND)


@app.route('/recent', methods=['GET'])
def recent():
    component_run_ids = get_recent_run_ids(
        request.args['limit']) if 'limit' in request.args else get_recent_run_ids()
    return json.dumps(component_run_ids)


@app.route('/trace', methods=['GET'])
def trace():
    if 'output_id' not in request.args:
        return error(f'output_id not specified.', HTTPStatus.NOT_FOUND)
    output_id = request.args['output_id']
    try:
        res = web_trace(output_id)
        return json.dumps(res)
    except RuntimeError:
        return error(f'Output {output_id} not found', HTTPStatus.NOT_FOUND)



from abc import ABC, abstractmethod


class Base(ABC):
    """ The Base class provides methods to save and store attributes as well as
    print them out."""

    # Taken from https://dev.to/mattconway1984/python-creating-instance-properties-2ej0
    def __setattr__(self, attr, value):
        try:
            # Try invoking the descriptor protocol __set__ "manually"
            got_attr = super().__getattribute__(attr)
            got_attr.__set__(self, value)
        except AttributeError:
            # Attribute is not a descriptor, just set it:
            super().__setattr__(attr, value)

    def __getattribute__(self, attr):
        # If the attribute does not exist, super().__getattribute__()
        # will raise an AttributeError
        got_attr = super().__getattribute__(attr)
        try:
            # Try "manually" invoking the descriptor protocol __get__()
            return got_attr.__get__(self, type(self))
        except AttributeError:
            # Attribute is not a descriptor, just return it:
            return got_attr

    def __iter__(self):
        for k in self._properties():
            yield k, self.__getattribute__(k)

    @classmethod
    def _properties(cls):
        return [p for p in cls.__dict__ if isinstance(getattr(cls, p), property)]

    @classmethod
    def from_dictionary(cls, d):
        d = {key: value for key,
             value in d.items() if key in cls._properties()}
        return cls(**d)

    def to_dictionary(self):
        return {k: self.__getattribute__(k) for k in self._properties()}

    @abstractmethod
    def __repr__(self):
        pass



from mltrace.db.utils import _create_engine_wrapper, _initialize_db_tables, _drop_everything, _map_extension_to_enum
from mltrace.db import Component, ComponentRun, IOPointer, PointerTypeEnum, Tag
from sqlalchemy import func
from sqlalchemy.orm import sessionmaker, joinedload

import logging
import typing


class Store(object):
    """Helper methods to interact with the db."""

    def __init__(self, uri: str, delete_first=False):
        """
        Creates the postgres database for the store. Raises exception if uri
        isn't prefixed with postgresql://.

        Args:
            uri (str): URI string to connect to the SQLAlchemy database.
        """
        if not uri.startswith('postgresql://'):
            raise RuntimeError(
                'Database URI must be prefixed with `postgresql://`')
        self.engine = _create_engine_wrapper(uri)

        # TODO(shreyashankar) remove this line
        if delete_first:
            _drop_everything(self.engine)

        # TODO(shreyashankar) check existing tables against expected tables
        _initialize_db_tables(self.engine)

        # Initialize session
        self.Session = sessionmaker(self.engine)
        self.session = self.Session()

    def __del__(self):
        """On destruction, close session."""
        self.session.close()

    def create_component(self, name: str, description: str, owner: str, tags: typing.List[str]):
        """Creates a component entity in the database if it does not already exist."""
        res = self.get_component(name)

        if res:
            logging.info(f'Component with name "{name}" already exists.')
            return

        # Add to the DB if it is not already there
        logging.info(
            f'Creating new Component with name "{name}", description "{description}", owner "{owner}", and tags "{tags}".')
        tags = [Tag(t) for t in tags]
        component = Component(
            name=name, description=description, owner=owner, tags=tags)
        self.session.add(component)
        self.session.commit()

    def get_component(self, name: str) -> Component:
        """Retrieves component if exists."""
        return self.session.query(Component).outerjoin(Tag, Component.tags).filter(Component.name == name).first()

    def get_component_run(self, id: str) -> ComponentRun:
        """Retrieves component run if exists."""
        return self.session.query(ComponentRun).filter(
            ComponentRun.id == id).first()

    def add_tags_to_component(self, component_name: str, tags: typing.List[str]):
        """Retreives existing component and adds tags."""
        component = self.get_component(component_name)

        if not component:
            raise RuntimeError(
                f'Component with name "{component_name}" not found.')

        tag_objects = [self.get_tag(t) for t in tags]
        component.add_tags(tag_objects)
        self.session.commit()

    def initialize_empty_component_run(self, component_name: str) -> ComponentRun:
        """Initializes an empty run for the specified component. Does not
        commit to the database."""
        component_run = ComponentRun(component_name=component_name)
        return component_run

    def get_tag(self, name=str) -> Tag:
        """Creates a tag around the name if it doesn't already exist."""
        res = self.session.query(Tag).filter(Tag.name == name).all()

        # Must create new Tag
        if len(res) == 0:
            logging.info(f'Creating new Tag with name "{name}".')
            tag = Tag(name)
            self.session.add(tag)
            self.session.commit()
            return tag

        # Return existing Tag
        return res[0]

    def get_io_pointer(self, name=str, pointer_type: PointerTypeEnum = None, create=True) -> IOPointer:
        """ Creates an io pointer around the specified path.
        Retrieves existing io pointer if exists in DB,
        otherwise creates a new one if create flag is set."""
        res = self.session.query(IOPointer).filter(IOPointer.name == name).all()

        # Must create new IOPointer
        if len(res) == 0 and create == True:
            logging.info(f'Creating new IOPointer with name "{name}".')
            if pointer_type == None:
                pointer_type = _map_extension_to_enum(name)

            iop = IOPointer(name=name, pointer_type=pointer_type)
            self.session.add(iop)
            self.session.commit()
            return iop

        # Return existing object
        return res[0]

    def delete_component(self, component: Component):
        self.session.delete(component)
        logging.info(
            f'Successfully deleted Component with name "{component.name}".')

    def delete_component_run(self, component_run: ComponentRun):
        self.session.delete(component_run)
        logging.info(
            f'Successfully deleted ComponentRun with id "{component_run.id}" and name "{component_run.component_name}".')

    def delete_io_pointer(self, io_pointer: IOPointer):
        self.session.delete(io_pointer)
        logging.info(
            f'Successfully deleted IOPointer with name "{io_pointer.name}".')

    def commit_component_run(self, component_run: ComponentRun):
        """Commits a fully initialized component run to the DB."""
        status_dict = component_run.check_completeness()
        if not status_dict['success']:
            raise RuntimeError(status_dict['msg'])

        # Commit to DB
        self.session.add(component_run)
        logging.info(
            f'Committing ComponentRun of type "{component_run.component_name}" to the database.')
        self.session.commit()

    def set_dependencies_from_inputs(self, component_run: ComponentRun):
        """ Gets IOPointers associated with component_run's inputs, checks
        against any ComponentRun's outputs, and if there are any matches,
        sets the ComponentRun's dependency on the most recent match."""
        input_ids = [inp.name for inp in component_run.inputs]
        matches = self.session.query(ComponentRun, func.max(ComponentRun.start_timestamp).over(partition_by=ComponentRun.component_name))\
            .outerjoin(IOPointer, ComponentRun.outputs).filter(
            IOPointer.name.in_(input_ids)).all()
        matches = [m[0] for m in matches]

        # If there are no matches, return
        if len(matches) == 0:
            return

        # Get match with the max timestamp and set upstream
        component_run.set_upstream(matches)

    def _traverse(self, node: ComponentRun, depth: int, node_list: typing.List[ComponentRun]):
        # Add node to node_list as the step
        node_list.append((depth, node))

        # Base case
        if len(node.dependencies) == 0:
            return

        # Recurse on neighbors
        for neighbor in node.dependencies:
            self._traverse(neighbor, depth + 1, node_list)

    def _web_trace_helper(self, component_run_object: ComponentRun):
        """ Helper function that populates the dictionary of ComponentRuns for
        the web trace. Returns dictionary and counter."""
        res = {}
        res['id'] = f'componentrun_{component_run_object.id}'
        res['label'] = component_run_object.component_name
        res['hasCaret'] = True
        res['isExpanded'] = True
        res['childNodes'] = []

        for out in component_run_object.outputs:
            out_dict = {
                'id': f'iopointer_{out.name}',
                'label': out.name,
                'hasCaret': False,
                'parent': res['id']
            }

            # Settle on icon
            if out.pointer_type == PointerTypeEnum.DATA:
                out_dict['icon'] = 'database'
            elif out.pointer_type == PointerTypeEnum.MODEL:
                out_dict['icon'] = 'function'

            res['childNodes'].append(out_dict)

        for dep in component_run_object.dependencies:
            child_res = self._web_trace_helper(dep)
            res['childNodes'].append(child_res)

        return res

    def web_trace(self, output_id: str):
        """Prints list of ComponentRuns to display in the UI."""
        component_run_objects = self.session.query(ComponentRun).outerjoin(
            IOPointer, ComponentRun.outputs).order_by(
            ComponentRun.start_timestamp.desc()).filter(IOPointer.name == output_id).all()

        if len(component_run_objects) == 0:
            raise RuntimeError(f'ID {output_id} does not exist.')

        return [self._web_trace_helper(cr) for cr in component_run_objects]

    def trace(self, output_id: str):
        """Prints trace for an output id.
        Returns list of tuples (level, ComponentRun) where level is how
        many hops away the node is from the node that produced the output_id."""
        component_run_object = self.session.query(ComponentRun).outerjoin(
            IOPointer, ComponentRun.outputs).filter(IOPointer.name == output_id).first()

        if component_run_object is None:
            raise RuntimeError(f'ID {output_id} does not exist.')

        print(f'Printing trace for output {output_id}...')

        node_list = []
        self._traverse(component_run_object, 1, node_list)
        return node_list

    def trace_batch(self, output_ids: typing.List[str]):
        pass

    def get_history(self, component_name: str, limit: int = 10) -> typing.List[ComponentRun]:
        """Gets lineage for the component, or a history of all its runs."""
        history = self.session.query(ComponentRun).filter(ComponentRun.component_name == component_name).order_by(
            ComponentRun.start_timestamp.desc()).limit(limit).all()

        return history

    def get_components_with_owner(self, owner: str) -> typing.List[Component]:
        """ Returns a list of all the components associated with the specified
        order."""
        components = self.session.query(Component).filter(
            Component.owner == owner).options(joinedload('tags')).all()

        if len(components) == 0:
            raise RuntimeError(f'Owner {owner} has no components.')

        return components

    def get_components_with_tag(self, tag: str) -> typing.List[Component]:
        """Returns a list of all the components associated with that tag."""
        components = self.session.query(Component).join(
            Tag, Component.tags).filter(Tag.name == tag).all()

        if len(components) == 0:
            raise RuntimeError(f'Tag {tag} has no components associated.')

        return components

    def get_recent_run_ids(self, limit: int = 50) -> typing.List[str]:
        """Returns a list of recent component run IDs."""
        runs = list(map(lambda x: int(x[0]), self.session.query(ComponentRun.id).order_by(
            ComponentRun.start_timestamp.desc()).limit(limit).all()))

        return runs



from sqlalchemy.ext.declarative import declarative_base


def _todict(obj: object) -> dict:
    """ Return the object's dict excluding private attributes, 
    sqlalchemy state and relationship attributes.
    """
    excl = ('_sa_adapter', '_sa_instance_state')
    return {k: v for k, v in vars(obj).items() if not k.startswith('_') and
            not any(hasattr(v, a) for a in excl)}


class BaseWithRepr:

    def __repr__(self):
        params = ', '.join(f'{k}={v}' for k, v in _todict(self).items())
        return f"{self.__class__.__name__}({params})"


Base = declarative_base(cls=BaseWithRepr)



from mltrace.entities.component import Component
from mltrace.entities.component_run import ComponentRun
from mltrace.entities.io_pointer import IOPointer

__all__ = [
    "Component",
    "ComponentRun",
    "IOPointer"
]



from mltrace.client import clean_db, create_component, register, backtrace, get_history, get_components_with_owner, tag_component, get_components_with_tag, log_component_run, create_random_ids, get_component_information, get_component_run_information, web_trace, get_recent_run_ids, get_io_pointer, set_db_uri, get_db_uri

__all__ = [
    "clean_db",
    "create_component",
    "register",
    "backtrace",
    "get_history",
    "get_components_with_owner",
    'tag_component',
    'get_components_with_tag',
    'log_component_run',
    'create_random_ids',
    'get_component_information',
    'get_component_run_information',
    'web_trace',
    'get_recent_run_ids',
    'get_io_pointer',
    'set_db_uri',
    'get_db_uri'
]



from mltrace.db.models import IOPointer, Component, ComponentRun, PointerTypeEnum, Tag
from mltrace.db.store import Store

__all__ = [
    "Component",
    "ComponentRun",
    "IOPointer",
    "Store",
    "PointerTypeEnum",
    "Tag"
]



from setuptools import setup, find_packages

import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name='mltrace',
    version='0.1',
    description='Lineage and tracing for ML pipelines',
    long_description=README,
    long_description_content_type="text/markdown",
    author='shreyashankar',
    author_email='shreya@cs.stanford.edu',
    license="Apache License 2.0",
    classifiers=["Intended Audience :: Developers",
                 "Programming Language :: Python", "Topic :: Software Development :: Bug Tracking"],
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'click',
        'docker',
        'flask',
        'furo',
        'gitpython',
        'psycopg2-binary',
        'pytest',
        'python-dotenv',
        'sqlalchemy'
    ],
    include_package_data=True,
    project_urls={
        "Bug Tracker": "https://github.com/loglabs/mltrace/issues",
        "Source Code": "https://github.com/loglabs/mltrace",
    },
)
